from encoders.utils import *
from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

import common
import random


class IncrementalWeakeningStrategyExtractor(object):

    def __init__(self, quantelim, environment):
        self.quantelim = quantelim
        
        self.environment = environment
        self.solver = MathSAT5Solver(environment)
        return


    def sortYVars(self, yvars):
        # Heuristic implementation, for now IDENTITY is enough
        return yvars


    def extractStrategy(self, clausifier):
        st = self.environment.symbolTable

        yvars = clausifier.yvars
        xvars = clausifier.xvars

        print("Starting strategy extraction...")

        print("  Sorting y variables...")
        yvars = self.sortYVars(yvars)

        print("  Preparing coefficients structures...")
        coeff = []
        for _ in range(0, len(xvars)):
            coeff.append([])

        for i in range(0, len(yvars) + 1):
            print("  Searching a strategy depending on the first %d uncontrollable variables..." % i)

            print("  Declaring new coefficients...")
            for k in range(0, len(xvars)):
                row = coeff[k]
                nv = st.createNewFreshSymbol(Types.REAL, "_C_%d")
                row.append(nv)
                self.solver.declareVariable(nv)

            print("  Encoding problem as SMT formula...")
            formula = self.encode(clausifier, coeff, i-1)
    
            print("  Resetting assertions...")
            self.solver.resetAssertions()

            print("  Asserting SMT formula...")
            self.solver.addAssertion(formula)
    
            print("  Solving SMT formula...")
            if self.solver.solve():
                print("  Extracting model...")
                s = LinearStrategy(yvars[:i])
                for i in range(0, len(xvars)):
                    row = coeff[i]
                    x = xvars[i]

                    vals = []
                    for c in row:
                        v = self.solver.getModelValue(c)
                        vals.append(v)
                    s.addPolicy(x, vals)
                print("Done.")
                return s

            else:
                print("  No strategy found under these observability assumptions.\n")

        print("No linear strategy exists!")
        print("Done.")
        return None


    def encode(self, clausifier, coeff, index):
        xvars = clausifier.xvars
        yvars = clausifier.yvars
        ycount = len(yvars)

        print("      Preparing formula definitions...")

        gamma = clausifier.getGamma()
        quantified = set(yvars[(index+1):])
        conj = []
        for c in clausifier.getPsiClauses():
            formula = Exists(quantified, And([Not(c), gamma]))
            elim = self.quantelim.removeQuantifiers(formula)
            conj.append(Not(elim))
            
        formula = And(conj)

        print("      Building problem formula...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            f = row[-1]
            for j in range(0, len(row) - 1):
                c = row[j]
                f = Plus(Times(c, yvars[j]), f)

            formula = Definition(x.name, f, formula)

        if index >= 0:
            resList = self.allCases(clausifier, list(yvars[:(index+1)]), formula)
            return And(resList)
        else:
            return formula


    def allCases(self, clausifier, choices, base, count=0, total=None):
        if len(choices) == 0:
            return [base]

        pivot = choices.pop()
        (l,u) = clausifier.getYBound(pivot)

        if total is None:
            total = len(choices)
        count = count+1
        
        sonList = self.allCases(clausifier, choices, base, count, total)

        perc = 100 / (2**count)

        print("        [%d/%d]\tGenerating cases for variable %s.\t\t(%.2f%% overall completion)" % (total-count+2, total+1, pivot.name, perc))
        res = []
        for s in sonList:
            res.append(Definition(pivot.name, Constant(l), s))
            res.append(Definition(pivot.name, Constant(u), s))

        return res

    pass
